{¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤@
@                                                                      @
@              Nom du produit : GVTransitions - transitions            @
@              Nom du fichier : transitions.pas                        @
@              Copyright : G. Vasseur 2018                             @
@              Contact : gillesvasseur58@gmail.com                     @
@              Date/Heure : 19/06/2018 09:10:19                        @
@              Version : 1.0.3.67                                      @
@                                                                      @
@¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤¤}

{ Transitions entre deux images - transitions

Copyright (C) 19/06/2018 09:10:19 G. Vasseur 2018 gillesvasseur58@gmail.com

This source is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your option)
any later version.

This code is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

A copy of the GNU General Public License is available on the World Wide Web
at <http://www.gnu.org/copyleft/gpl.html>. You can also obtain it by writing
to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA.
}

unit transitions;

{$I defines.inc}

interface

uses
  Classes, SysUtils, Controls, ExtCtrls, Graphics, Types,
  BGRABitmapTypes, BGRABitmap,
  transitionstypes;

type

  TGVTransitionStateEvent = procedure(Sender: TObject; State: TGVTransitionState;
    Trans: TGVTransitionType; Step: Integer) of object;

  { TGVTransition }

  TGVTransition = class(TGraphicControl)
  strict private
    fQuality: TGVQuality;
    fSmooth: TGVSmooth;
    fSpecialEffect: TGVSpecialEffect;
    fTimer: TTimer;
    fLBGRAFrom, fLBGRATo, fLBGRAMask: TBGRABitmap;
    fY, fX: Single;
    fPts, fPtsTemp: array of TPointF;
    fAutoStrips: boolean;
    fPictureFrom, fPictureTo: TPicture;
    fBGRAFrom, fBGRATo: TBGRABitmap;
    fHorizontalStrips, fVerticalStrips: Integer;
    fInterpolation: TGVInterpolation;
    fTransitionChange, fLooping: TNotifyEvent;
    fStateChanged: TGVTransitionStateEvent;
    fOpacityState: TGVOpacityState;
    fSpeed: Integer;
    fState: TGVTransitionState;
    fTransition: TGVTransitionType;
    fStep: Integer;
    fInc: Integer;
    fLastLoop: Integer;
    function GetPictureFrom: TPicture;
    function GetPictureTo: TPicture;
    function GetVersion: string;
    procedure SetAutoStrips(AValue: boolean);
    procedure SetHorizontalStrips(AValue: Integer);
    procedure SetInterpolation(AValue: TGVInterpolation);
    procedure SetOpacityState(AValue: TGVOpacityState);
    procedure SetPictureFrom(AValue: TPicture);
    procedure SetPictureTo(AValue: TPicture);
    procedure SetQuality(AValue: TGVQuality);
    procedure SetSmooth(AValue: TGVSmooth);
    procedure SetSpecialEffect(AValue: TGVSpecialEffect);
    procedure SetSpeed(AValue: Integer);
    procedure SetState(AValue: TGVTransitionState);
    procedure SetTransition(AValue: TGVTransitionType);
    procedure SetVerticalStrips(AValue: Integer);
    function DestinationOpacity: Integer; inline;
    function SourceOpacity: Integer; inline;
  protected
    function ComputeInterpolation(AStart, AEnd: Single): Single; virtual;
    function ComputeInterpolationInt(AStart, AEnd: Single): Integer;
    procedure Paint; override;
    procedure InitTransition; virtual;
    procedure ComputeSource; virtual;
    procedure ComputeDestination; virtual;
    procedure ComputeFinalDraw; virtual;
    procedure DoLoop; virtual;
    procedure Change; virtual;
    procedure StateChanged; virtual;
    procedure Looping; virtual;
    procedure DoReset; virtual;
    procedure Resize; override;
    procedure OnTimerEvent(Sender: TObject);
    class function GetControlClassDefaultSize: TSize; override;

    // liste des transitions
    procedure DoNegateMask;
    procedure DoPrepareSpline;
    procedure DoPrepareSplineUpOrDown;
    procedure DoUp;
    procedure DoDown;
    procedure DoRight;
    procedure DoLeft;
    procedure DoBottomRight;
    procedure DoBottomLeft;
    procedure DoTopRight;
    procedure DoTopLeft;
    procedure DoUncoverUpOrDown(Up: boolean);
    procedure DoUncover(ToRight: boolean);
    procedure DoOverUpOrDown(Up: boolean);
    procedure DoOver(ToRight: boolean);
    procedure DoUncoverBottom(ToRight: boolean);
    procedure DoUncoverTop(ToRight: boolean);
    procedure DoUncoverHorizontalBackAndForth;
    procedure DoUncoverVerticalBackAndForth;
    procedure DoOverSlipTop(ToRight: boolean);
    procedure DoOverSlipBottom(ToRight: boolean);
    procedure DoOverBottom(ToRight: boolean);
    procedure DoOverTop(ToRight: boolean);
    procedure DoBottomExpand(ToRight: boolean);
    procedure DoTopExpand(ToRight: boolean);
    procedure DoLeaveUpOrDown(Up: boolean);
    procedure DoLeave(ToRight: boolean);
    procedure DoLeaveBottom(ToRight: boolean);
    procedure DoLeaveTop(ToRight: boolean);
    procedure DoVerticalExpand;
    procedure DoHorizontalExpand;
    procedure DoHorizontalShrink;
    procedure DoVerticalShrink;
    procedure DoRectExpand;
    procedure DoRectShrink;
    procedure DoEllipseExpand;
    procedure DoEllipseShrink;
    procedure DoDiskExpand;
    procedure DoDiskShrink;
    procedure DoCrossExpand;
    procedure DoCrossShrink;
    procedure DoShrinkExpandUpOrDown(Up: boolean);
    procedure DoTrianglesExpand;
    procedure DoTrianglesShrink;
    procedure DoMegaStar;
    procedure DoSplinesDown;
    procedure DoSplinesUp;
    procedure DoSplinesRight;
    procedure DoSplinesLeft;
    procedure DoStrips(ToRight: boolean);
    procedure DoStripsUpOrDown(Up: boolean);
    procedure DoStripsHorizontalInterlaced(LeftFirst: boolean);
    procedure DoStripsInterlaced(UpFirst: boolean);
    procedure DoRectUpOrDown(Up: Boolean);
    procedure DoWipe(ToRight: boolean);
    procedure DoOverSin(ToRight: boolean);
    function Parabola(AOffset, AMax, APerCent: Integer): Single;
    procedure DoBouncing(ToRight: boolean);
    procedure DoJump(ToRight: boolean);
    procedure DoBlank;
    procedure DoFadeIn;
    procedure DoRotateRight(CCW: boolean);
    procedure DoCenterRotate(CCW: boolean);
    procedure DoRotateLeft(CCW: boolean);
    procedure DoSpiral(CCW: Boolean);
    procedure DoGrowCenter;
    procedure DoGrowBottom(ToRight: Boolean);
    procedure DoGrowTop(ToRight: Boolean);
    procedure DoGrowRotate(CCW: Boolean);
    procedure DoGrowCenterRotate(CCW: Boolean);
    procedure DoGrowOver(ToRight: Boolean);
    procedure DoGrowOverUpOrDown(Up: Boolean);
    procedure DoRotateGrowLeft(CCW:Boolean);
    procedure DoRotateGrowTopLeft(CCW:Boolean);
    procedure DoRotateTopLeft(CCW:Boolean);
    procedure DoRotateBottomRight(CCW:Boolean);
    procedure DoMappingDown;
    procedure DoMappingUp;
    procedure DoMultiTwirl;
    procedure DoTwirl;
    procedure DoSphere;
    procedure DoCylinderRight;
    procedure DoCylinderLeft;
    procedure DoMultiExpand;
    procedure DoMultiExpandTwirl;
    procedure DoDrawFromBounds;
    procedure DoOffset(ToRight: Boolean);
    procedure DoPixelate;
    procedure DoBlurMotion;
    {$IFDEF Debug_Transitions}
    procedure DoInitDummy;
    function DoSourceDummy: TBGRABitmap;
    procedure DoDestinationDummy;
    procedure DoFinalDummy;
    {$ENDIF}

  public
    // copyright du composant
    class function Copyright: string; static;
    // créatiton du composant
    constructor Create(AOwner: TComponent); override;
    // destruction du composant
    destructor Destroy; override;
    // nombre de transitions
    function Count: Integer;
    // opacité en cours (de 0 à 255)
    function GetOpacity(Up: Boolean = True): Integer;
    // calcul du nombre de bandes pour ASize
    function NumberOfStrips(ASize: Integer): Integer;
    // démarrage de la transition
    procedure Start;
    // arrêt de la transition
    procedure Stop;
    // remise à zéro du composant
    procedure Reset;
    // suspension de la transition en cours (si elle existe)
    procedure Suspend;
    // reprise de la transition suspendue (si elle existe)
    procedure Resume;
    // indicateur de transition en cours
    function IsRunning: Boolean;
    // étape en cours
    property Step: Integer read fStep;
    // état en cours de la transition
    property State: TGVTransitionState read fState write SetState default trsWaiting;

  published
    // propriétés héritées
    property Color;
    property ParentColor;
    property Align;
    property Anchors;
    property AutoSize;
    property BorderSpacing;
    property Constraints;
    property DragCursor;
    property DragMode;
    property Enabled;
    property OnChangeBounds;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseEnter;
    property OnMouseLeave;
    property OnMouseMove;
    property OnMouseUp;
    property OnMouseWheel;
    property OnMouseWheelDown;
    property OnMouseWheelUp;
    property OnPaint;
    property OnResize;
    property OnStartDrag;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Visible;

    // version du composant
    property Version: string read GetVersion;
    // image d'origine
    property PictureFrom: TPicture read GetPictureFrom write SetPictureFrom;
    // image de destination
    property PictureTo: TPicture read GetPictureTo write SetPictureTo;
    // interpolation d'image à image
    property Interpolation: TGVInterpolation read fInterpolation
      write SetInterpolation default intLinear;
    // vitesse de transition
    property Speed: Integer read fSpeed write SetSpeed default C_DefaultSpeed;
    // type d'opacité utilisée
    property Opacity: TGVOpacityState
      read fOpacityState write SetOpacityState default opaBoth;
    // type de transition
    property Transition: TGVTransitionType
      read fTransition write SetTransition default traNone;
    // mode automatique de calcul du nombre de bandes
    property AutoStrips: Boolean read fAutoStrips write SetAutoStrips;
    // nombre de bandes horizontales (modifiable si AutoStrips est à False)
    property HorizontalStrips: Integer read fHorizontalStrips
      write SetHorizontalStrips;
    // nombre de bandes verticales (modifiable si AutoStrips est à False)
    property VerticalStrips: Integer read fVerticalStrips write SetVerticalStrips;
    // qualité des transformations (homothéties)
    property Quality: TGVQuality read fQuality write SetQuality default quaNormal;
    // effets spéciaux
    property SpecialEffect: TGVSpecialEffect read fSpecialEffect write SetSpecialEffect default speNone;
    // lissage des transitions
    property Smooth: TGVSmooth read fSmooth write SetSmooth default smoMedium;
    // le composant a changé - gestionnaire de l'événement
    property OnChange: TNotifyEvent read fTransitionChange
      write fTransitionChange;
    // l'état de la transition a changé - gestionnaire de l'événement
    property OnStateChange: TGVTransitionStateEvent read fStateChanged
      write fStatechanged;
    // étape suivante - gestionnaire de l'événement
    property OnLoop: TNotifyEvent read fLooping write fLooping;
  end;

procedure Register;


implementation

uses
  Math, LResources, BGRATransform;

procedure Register;
begin
  RegisterComponents('gvsoft', [TGVTransition]);
end;

{ TTransition }

procedure TGVTransition.SetInterpolation(AValue: TGVInterpolation);
// *** détermination de l'interpolation ***
begin
  if fInterpolation = AValue then
    exit;
  fInterpolation := AValue;
  Change;
end;

procedure TGVTransition.SetOpacityState(AValue: TGVOpacityState);
// *** détermination de l'opacité ***
begin
  if fOpacityState = AValue then
    exit;
  fOpacityState := AValue;
  Change;
end;

function TGVTransition.GetPictureFrom: TPicture;
// *** restitution de l'image d'origine ***
begin
  Result := fPictureFrom;
end;

function TGVTransition.GetPictureTo: TPicture;
// *** restitution de l'image de destination ***
begin
  Result := fPictureTo;
end;

function TGVTransition.GetVersion: string;
// *** version du composant ***
begin
  Result := C_GVVersionSt;
end;

procedure TGVTransition.SetAutoStrips(AValue: boolean);
// *** calcul automatique du nombre de bandes ***
begin
  if fAutoStrips = AValue then
    exit;
  fAutoStrips := AValue;
  if fAutoStrips then
  begin
    fHorizontalStrips := NumberOfStrips(ClientHeight);
    fVerticalStrips := NumberOfStrips(ClientWidth);
  end;
end;

procedure TGVTransition.SetHorizontalStrips(AValue: Integer);
// *** nombre de bandes horizontales ***
begin
  if (fHorizontalStrips = AValue) or AutoStrips then
    exit;
  fHorizontalStrips := Max(AValue, C_MinHorizontalStrips);
  Change;
end;

procedure TGVTransition.SetPictureFrom(AValue: TPicture);
// *** acquisition de l'image d'origine ***
begin
  fPictureFrom.Assign(AValue);
  FreeAndNil(fBGRAFrom);
  fBGRAFrom := TBGRABitmap.Create(AValue.Bitmap);
  BGRAReplace(fBGRAFrom, fBGRAFrom.Resample(ClientWidth, ClientHeight));
  Change;
end;

procedure TGVTransition.SetPictureTo(AValue: TPicture);
// *** acquisition de l'image de destination ***
begin
  fPictureTo.Assign(AValue);
  FreeAndNil(fBGRATo);
  fBGRATo := TBGRABitmap.Create(AValue.Bitmap);
  BGRAReplace(fBGRATo, fBGRATo.Resample(ClientWidth, ClientHeight));
  Change;
end;

procedure TGVTransition.SetQuality(AValue: TGVQuality);
// *** qualité des transformations (homothéties) ***
begin
  if fQuality = AValue then
    exit;
  fQuality := AValue;
  case fQuality of
    quaLow: fLBGRAFrom.ResampleFilter := rfBox;
    quaNormal: fLBGRAFrom.ResampleFilter := rfHalfCosine;
    quaBest: fLBGRAFrom.ResampleFilter := rfBestQuality;
  end;
end;

procedure TGVTransition.SetSmooth(AValue: TGVSmooth);
// *** lissage des transitions ***
begin
  if fSmooth = AValue then
    exit;
  fSmooth := AValue;
  case fSmooth of
    smoHigh: begin
      fInc := C_DefaultInc div 2;
      fLastLoop := C_DefaultLastLoop;
    end;
    smoMedium: begin
      fInc := C_DefaultInc;
      fLastLoop := C_DefaultLastLoop;
    end;
    smoNone, smoLow: begin
      fInc := C_DefaultInc;
      fLastLoop := C_DefaultLastLoop div 2;
    end;
    smoFast: begin
      fInc := C_DefaultInc * 2;
      fLastLoop := C_DefaultLastLoop div 2;
    end;
  end;
  Change;
end;

procedure TGVTransition.SetSpecialEffect(AValue: TGVSpecialEffect);
// *** détermination des effets spéciaux ***
begin
  if fSpecialEffect = AValue then
    exit;
  fSpecialEffect := AValue;
  Change;
end;

procedure TGVTransition.SetSpeed(AValue: Integer);
// *** détermination de la vitesse ***
begin
  if fSpeed = AValue then
    exit;
  fSpeed := AValue;
  fTimer.Interval := C_DefaultTimer - (fSpeed div 2);
  Change;
end;

procedure TGVTransition.SetState(AValue: TGVTransitionState);
// *** changement d'état ***
begin
  if fState = AValue then
    exit;
  fState := AValue;
  if Assigned(fTimer) then
    fTimer.Enabled := (fState = trsRunning);
  StateChanged;
end;

procedure TGVTransition.SetTransition(AValue: TGVTransitionType);
// *** détermination de la transition active ***
begin
  if (fTransition = AValue) or (fState = trsRunning) then
    exit;
  fTransition := AValue;
  Change;
end;

procedure TGVTransition.SetVerticalStrips(AValue: Integer);
// *** nombre de bandes verticales ***
begin
  if (fVerticalStrips = AValue) or AutoStrips then
    exit;
  fVerticalStrips := Max(AValue, C_MinVerticalStrips);
  Change;
end;

function TGVTransition.DestinationOpacity: Integer;
// *** opacité pour l'image de destination ***
begin
  Result := ifthen(fOpacityState in [opaDestination, opaBoth], GetOpacity, C_MaxOpacity);
end;

function TGVTransition.SourceOpacity: Integer;
// *** opacité de l'image d'origine ***
begin
  Result := ifthen(fOpacityState in [opaSource, opaBoth], GetOpacity(False),
    C_MaxOpacity);
end;

procedure TGVTransition.Paint;
// *** dessin ***
var
  LX, LY: Integer;
begin
  if csDesigning in ComponentState then
  begin
    with Canvas do
    begin
      Font.Color := clBlue;
      Font.Size := 16;
      LX := (Self.Width - TextWidth(Self.Name)) div 2;
      if LX < 0 then
        LX := 0;
      LY := (Self.Height - TextHeight(Self.Name)) div 2;
      if LY < 0 then
        LY := 0;
      TextOut(LX, LY, Self.Name);
    end;
    exit;
  end;
  if Assigned(OnPaint) then
  begin
    Canvas.Font := Font;
    Canvas.Brush.Color := Color;
    inherited Paint;
  end;
  if Assigned(fLBGRAFrom) then
    fLBGRAFrom.Draw(Canvas, 0, 0);
end;

constructor TGVTransition.Create(AOwner: TComponent);
// *** construction du composant ***
begin
  inherited Create(AOwner);
  with GetControlClassDefaultSize do
    SetInitialBounds(0, 0, CX, CY);
  fInterpolation := intLinear;
  fSpeed := C_DefaultSpeed;
  fTransition := traNone;
  fOpacityState := opaBoth;
  fSpecialEffect := speNone;
  fTimer := TTimer.Create(Self);
  fTimer.Enabled := False;
  fTimer.Interval := C_DefaultTimer - (fSpeed div 2);
  fTimer.OnTimer := @OnTimerEvent;
  fPictureFrom := TPicture.Create;
  fPictureTo := TPicture.Create;
  fBGRAFrom := TBGRABitmap.Create(ClientWidth, ClientHeight, BGRABlack);
  fBGRATo := fBGRAFrom.Duplicate as TBGRABitmap;
  fLBGRAFrom := fBGRAFrom.Duplicate as TBGRABitmap;
  fLBGRATo := fBGRAFrom.Duplicate as TBGRABitmap;
  fLBGRAMask := fBGRAFrom.Duplicate as TBGRABitmap;
  fHorizontalStrips := NumberOfStrips(ClientHeight);
  fVerticalStrips := NumberOfStrips(ClientWidth);
  fState := trsWaiting;
  fQuality := quaNormal;
  fSmooth := smoMedium;
  fInc := C_DefaultInc;
  fLastLoop := C_DefaultLastLoop;
end;

destructor TGVTransition.Destroy;
// *** destruction du composant ***
begin
  fBGRAFrom.Free;
  fBGRATo.Free;
  fLBGRAFrom.Free;
  fLBGRATo.Free;
  fLBGRAMask.Free;
  fPictureTo.Free;
  fPictureFrom.Free;
  inherited Destroy;
end;

function TGVTransition.Count: Integer;
// *** nombre de transitions ***
begin
  Result := Ord(High(TGVTransitionType)) + 1;
end;

function TGVTransition.ComputeInterpolation(AStart, AEnd: Single): Single;
// *** calcul d'une interpolation ***
var
  LDiff: Single;
  LSign, LStep: Integer;

  function Exponant(AExp: byte): Single; inline;
  begin
    Result := LDiff * Power(LStep, AExp) / Power(100, AExp);
  end;

  function DownExponant(AExp: byte): Single; inline;
  begin
    Result := LDiff - LDiff * Power(100 - LStep, AExp) /
      Power(100, AExp);
  end;

begin
  LDiff := Abs(AEnd - AStart);
  if AStart > AEnd then
    LSign := -1
  else
    LSign := 1;
  LStep := fStep * 100 div fLastLoop;
  case Interpolation of
    intLinear: Result := Exponant(1);
    intQuadratic: Result := Exponant(2);
    intCubic: Result := Exponant(3);
    intQuartic: Result := Exponant(4);
    intQuintic: Result := Exponant(5);
    intSinus: Result := Exponant(1) * sin(pi * LStep / 200);
    intSlowDownQuadratic: Result := DownExponant(2);
    intSlowDownCubic: Result := DownExponant(3);
    intSlowDownQuartic: Result := DownExponant(4);
    intSlowDownQuintic: Result := DownExponant(5);
    intSpring: Result := Exponant(1) * Power(cos(pi * LStep / 100), 2);
    intExpo: Result := Exponant(1) * (Power(2, LStep / 100) - 1);
    intSqrt: Result := Exponant(1) * (Sqrt(LStep) / 10);
    intBounceCos: Result := Exponant(1) + (cos(LStep * pi / 100) + 1) * Exponant(1);
    intSlowDownCos: Result := Exponant(1) + (cos(LStep * pi / 100) + 1) * LStep;
    intStepsCos: Result := Exponant(1) + Power((cos(LStep * pi / 100) + 1), 2) * LStep;
    intCosine: Result := Exponant(1) * (1-cos(LStep/100*Pi))*0.5;
    intHalfCosine: Result := Exponant(1) * ((1-cos(LStep/100*Pi))*0.25 + LStep/100*0.5);
  end;
  Result := AStart + LSign * Result;
end;

function TGVTransition.ComputeInterpolationInt(AStart, AEnd: Single): Integer;
// *** calcul d'une interpolation (entiers) ***
begin
  Result := Round(ComputeInterpolation(AStart, AEnd));
end;

procedure TGVTransition.InitTransition;
// *** pré-traitement ***
begin
  case Transition of
    traSplinesLeft, traSplinesRight: DoPrepareSpline;
    traSplinesUp, traSplinesDown: DoPrepareSplineUpOrDown;
    {$IFDEF Debug_Transitions}
    traDummy: DoInitDummy;
    {$ENDIF}
  end;
end;

procedure TGVTransition.ComputeSource;
// *** traitement de l'image source ***
var
  LBGRATemp: TBGRABitmap;
begin
  fX := 0;
  fY := 0;
  fLBGRAFrom.Fill(Color);
  case Transition of
    traGrowPushUp, traLeaveUp, traPushUp: DoUp;
    traGrowPushDown, traLeaveDown, traPushDown: DoDown;
    traGrowPushLeft, traLeaveLeft, traPushLeft: DoLeft;
    traGrowPushRight, traLeaveRight, traPushRight: DoRight;
    traLeaveBottomRight: DoBottomRight;
    traLeaveBottomLeft: DoBottomLeft;
    traLeaveTopRight: DoTopRight;
    traLeaveTopLeft: DoTopLeft;
  end;
  // effets spéciaux
  case fSpecialEffect of
    speEmboss: LBGRATemp := fBGRAFrom.FilterEmboss(ComputeInterpolation(0, 180),
      ComputeInterpolationInt(64, 256)) as TBGRABitmap;
    speContour: LBGRATemp := fBGRAFrom.FilterContour as TBGRABitmap;
    speNegative: begin
      LBGRATemp := fBGRAFrom.Duplicate as TBGRABitmap;
      LBGRATemp.Negative;
    end;
    speLinearNegative: begin
      LBGRATemp := fBGRAFrom.Duplicate as TBGRABitmap;
      LBGRATemp.LinearNegative;
    end;
    speSwapRedBlue: begin
      LBGRATemp := fBGRAFrom.Duplicate as TBGRABitmap;
      LBGRATemp.SwapRedBlue;
    end;
    speSphere: LBGRATemp := fBGRAFrom.FilterSphere as TBGRABitmap;
    speRotate: LBGRATemp := fBGRAFrom.FilterRotate(PointF(ClientWidth / 2,
      ClientHeight/ 2), ComputeInterpolation(0, 360)) as TBGRABitmap;
    speGrayscale: begin
      LBGRATemp := fBGRAFrom.Duplicate as TBGRABitmap;
      LBGRATemp.InplaceGrayscale;
    end;
    spePixelate: LBGRATemp := fBGRAFrom.FilterPixelate(ComputeInterpolationInt(2, 24),
      True) as TBGRABitmap;
    speTwirl: LBGRATemp := fBGRAFrom.FilterTwirl(
      Point(ClientWidth div 2, ClientHeight div 2),
      ifthen(ClientWidth > ClientHeight, ComputeInterpolation(0, ClientWidth),
      ComputeInterpolation(0, ClientHeight))) as TBGRABitmap;
    speBlurMotion: LBGRATemp := fBGRAFrom.FilterBlurMotion(ComputeInterpolation(0, 15),
      ComputeInterpolation(0, 180), False) as TBGRABitmap;
  else
    LBGRATemp := fBGRAFrom.Duplicate as TBGRABitmap;
  end;
  {$IFDEF Debug_Transitions}
  if Transition = traDummy then
  begin
    FreeAndNil(LBGRATemp);
    LBGRATemp := DoSourceDummy;
  end;
  {$ENDIF}
  try
    fLBGRAFrom.PutImage(Round(fX), Round(fY), LBGRATemp, dmDrawWithTransparency,
      SourceOpacity);
  finally
    LBGRATemp.Free;
  end;
end;

procedure TGVTransition.ComputeDestination;
// *** traitement de l'image destination ***
begin
  fX := 0;
  fY := 0;
  fLBGRAMask.Fill(BGRABlack);
  case Transition of
    traUncoverDown: DoUncoverUpOrDown(False);
    traUncoverUp: DoUncoverUpOrDown(True);
    traUncoverLeft: DoUncover(False);
    traUncoverRight: DoUncover(True);
    traOverDown, traPushDown: DoOverUpOrDown(False);
    traOverUp, traPushUp: DoOverUpOrDown(True);
    traOverLeft, traPushLeft: DoOver(False);
    traOverRight, traPushRight: DoOver(True);
    traUncoverBottomRight: DoUncoverBottom(True);
    traUncoverBottomLeft: DoUncoverBottom(False);
    traUncoverTopRight: DoUncoverTop(True);
    traUncoverTopLeft: DoUncoverTop(False);
    traUncoverHorizontalBackAndForth: DoUncoverHorizontalBackAndForth;
    traUncoverVerticalBackAndForth: DoUncoverVerticalBackAndForth;
    traOverSlipBottomRight: DoOverSlipBottom(True);
    traOverSlipBottomLeft: DoOverSlipBottom(False);
    traOverSlipTopRight: DoOverSlipTop(True);
    traOverSlipTopLeft: DoOverSlipTop(False);
    traOverBottomRight: DoOverBottom(True);
    traOverBottomLeft: DoOverBottom(False);
    traOverTopRight: DoOverTop(True);
    traOverTopLeft: DoOverTop(False);
    traBottomRightExpand: DoBottomExpand(True);
    traBottomLeftExpand: DoBottomExpand(False);
    traTopRightExpand: DoTopExpand(True);
    traTopLeftExpand: DoTopExpand(False);
    traLeaveUp: DoLeaveUpOrDown(True);
    traLeaveDown: DoLeaveUpOrDown(False);
    traLeaveLeft: DoLeave(False);
    traLeaveRight: DoLeave(True);
    traBottomRightShrink, traLeaveBottomRight: DoLeaveBottom(True);
    traBottomLeftShrink, traLeaveBottomLeft: DoLeaveBottom(False);
    traTopRightShrink, traLeaveTopRight: DoLeaveTop(True);
    traTopLeftShrink, traLeaveTopLeft: DoLeaveTop(False);
    traHorizontalExpand: DoHorizontalExpand;
    traVerticalExpand: DoVerticalExpand;
    traHorizontalShrink: DoHorizontalShrink;
    traVerticalShrink: DoVerticalShrink;
    traRectExpand: DoRectExpand;
    traRectShrink: DoRectShrink;
    traEllipseExpand: DoEllipseExpand;
    traEllipseShrink: DoEllipseShrink;
    traDiskExpand: DoDiskExpand;
    traDiskShrink: DoDiskShrink;
    traCrossExpand: DoCrossExpand;
    traCrossShrink: DoCrossShrink;
    traShrinkExpandDown: DoShrinkExpandUpOrDown(False);
    traShrinkExpandUp: DoShrinkExpandUpOrDown(True);
    traTrianglesExpand: DoTrianglesExpand;
    traTrianglesShrink: DoTrianglesShrink;
    traMegaStar: DoMegaStar;
    traSplinesDown: DoSplinesDown;
    traSplinesUp: DoSplinesUp;
    traSplinesRight: DoSplinesRight;
    traSplinesLeft: DoSplinesLeft;
    traStripsDown: DoStripsUpOrDown(False);
    traStripsUp: DoStripsUpOrDown(True);
    traStripsRight: DoStrips(True);
    traStripsLeft: DoStrips(False);
    traStripsInterlacedLeftRight: DoStripsHorizontalInterlaced(True);
    traStripsInterlacedRightLeft: DoStripsHorizontalInterlaced(False);
    traStripsInterlacedUpDown: DoStripsInterlaced(True);
    traStripsInterlacedDownUp: DoStripsInterlaced(False);
    traRectUpDown: DoRectUpOrDown(False);
    traRectDownUp: DoRectUpOrDown(True);
    traWipeRight: DoWipe(True);
    traWipeLeft: DoWipe(False);
    traOverRightSin: DoOverSin(True);
    traOverLeftSin: DoOverSin(False);
    traBouncingRight: DoBouncing(True);
    traBouncingLeft: DoBouncing(False);
    traJumpRight: DoJump(True);
    traJumpLeft: DoJump(False);
    traFadeIn: DoFadeIn;
    {$IFDEF Debug_Transitions}
    traDummy: DoDestinationDummy;
    {$ENDIF}
  else
    DoNegateMask;
  end;
  fLBGRATo.PutImage(Round(fX), Round(fY), fBGRATo, dmSet);
  fLBGRATo.ApplyMask(fLBGRAMask);
end;

procedure TGVTransition.ComputeFinalDraw;
// *** traitement final de l'image ***
begin
  case Transition of
    traCCWRotateRight: DoRotateRight(True);
    traCWRotateRight: DoRotateRight(False);
    traCCWCenterRotate: DoCenterRotate(True);
    traCWCenterRotate: DoCenterRotate(False);
    traCCWRotateLeft: DoRotateLeft(True);
    traCWRotateLeft: DoRotateLeft(False);
    traCWSpiral: DoSpiral(False);
    traCCWSpiral: DoSpiral(True);
    traGrowCenter: DoGrowCenter;
    traGrowBottomRight: DoGrowBottom(True);
    traGrowBottomLeft: DoGrowBottom(False);
    traGrowTopRight: DoGrowTop(True);
    traGrowTopLeft: DoGrowTop(False);
    traGrowCWRotate: DoGrowRotate(False);
    traGrowCCWRotate: DoGrowRotate(True);
    traGrowCenterCCWRotate: DoGrowCenterRotate(True);
    traGrowCenterCWRotate: DoGrowCenterRotate(False);
    traGrowPushRight, traGrowOverRight: DoGrowOver(True);
    traGrowPushLeft, traGrowOverLeft: DoGrowOver(False);
    traGrowPushUp, traGrowOverUp: DoGrowOverUpOrDown(True);
    traGrowPushDown, traGrowOverDown: DoGrowOverUpOrDown(False);
    traRotateCCWGrowLeft: DoRotateGrowLeft(True);
    traRotateCWGrowLeft: DoRotateGrowLeft(False);
    traRotateCCWGrowTopLeft: DoRotateGrowTopLeft(True);
    traRotateCWGrowTopLeft: DoRotateGrowTopLeft(False);
    traRotateCCWTopLeft: DoRotateTopLeft(True);
    traRotateCWTopLeft: DoRotateTopLeft(False);
    traRotateCCWBottomRight: DoRotateBottomRight(True);
    traRotateCWBottomRight: DoRotateBottomRight(False);
    traMappingDown: DoMappingDown;
    traMappingUp: DoMappingUp;
    traMultiTwirl: DoMultiTwirl;
    traTwirl: DoTwirl;
    traSphere: DoSphere;
    traCylinderRight: DoCylinderRight;
    traCylinderLeft: DoCylinderLeft;
    traMultiExpand: DoMultiExpand;
    traMultiExpandTwirl: DoMultiExpandTwirl;
    traDrawFromBounds: DODrawFromBounds;
    traOffsetTopLeft: DoOffset(False);
    traOffsetBottomRight: DoOffset(True);
    traPixelate: DoPixelate;
    traBlurMotion: DoBlurMotion;
    {$IFDEF Debug_Transitions}
    traDummy: DoFinalDummy;
    {$ENDIF}
    else
      fLBGRAFrom.PutImage(0, 0, fLBGRATo, dmDrawWithTransparency, DestinationOpacity);
  end;
end;

{$I transitions.inc }

procedure TGVTransition.DoLoop;
// *** boucle de la transition ***
begin
  Inc(fStep, fInc);
  ComputeSource;
  ComputeDestination;
  ComputeFinalDraw;
  Looping;
  Repaint;
  if fStep >= fLastLoop then
  begin
    State := trsEnding;
    State := trsWaiting;
  end;
end;

procedure TGVTransition.Change;
// *** notification de changement d'état ***
begin
  if Assigned(fTransitionChange) then
    fTransitionChange(Self);
end;

procedure TGVTransition.StateChanged;
// *** notification de changement d'état ***
begin
  if Assigned(fStateChanged) then
    fStateChanged(Self, fState, fTransition, fStep);
end;

procedure TGVTransition.Looping;
// *** notification de l'étape suivante ***
begin
  if Assigned(fLooping) then
    fLooping(Self);
end;

procedure TGVTransition.DoReset;
// *** remise à zéro ***
begin
  if Assigned(fPictureFrom) and Assigned(fPictureTo) then
  begin
    if (fPictureFrom.Height > 0) and (fPictureFrom.Width > 0) and
      (fPictureTo.Height > 0) and (fPictureTo.Width > 0) then
    begin
      FreeAndNil(fBGRAFrom);
      fBGRAFrom := TBGRABitmap.Create(fPictureFrom.Bitmap);
      BGRAReplace(fBGRAFrom, fBGRAFrom.Resample(ClientWidth, ClientHeight));
      fLBGRAFrom.Assign(fBGRAFrom);
      FreeAndNil(fBGRATo);
      fBGRATo := TBGRABitmap.Create(fPictureTo.Bitmap);
      BGRAReplace(fBGRATo, fBGRATo.Resample(ClientWidth, ClientHeight));
      fLBGRATo.Assign(fBGRATo);
      FreeAndNil(fLBGRAMask);
      fLBGRAMask := TBGRABitmap.Create(ClientWidth, ClientHeight, BGRABlack);
      fStep := 0;
      Change;
    end;
  end;
end;

procedure TGVTransition.Resize;
// *** le contrôle est redimensionné ***
var
  LState: TGVTransitionState;
begin
  LState := fState;
  State := trsPaused;
  try
    inherited Resize;
    if AutoStrips then
    begin
      fHorizontalStrips := NumberOfStrips(ClientHeight);
      fVerticalStrips := NumberOfStrips(ClientWidth);
    end;
    if not (csDesigning in ComponentState) and not (fState = trsRunning) then
      DoReset;
  finally
    State := LState;
  end;
end;

procedure TGVTransition.OnTimerEvent(Sender: TObject);
// *** événement lié au timer ***
begin
  DoLoop;
end;

class function TGVTransition.GetControlClassDefaultSize: TSize;
// *** taille par défaut ***
begin
  Result := inherited GetControlClassDefaultSize;
  Result.cx := C_DefaultWidth;
  Result.cy := C_DefaultHeight;
end;

class function TGVTransition.Copyright: string;
// *** copyright ***
begin
  Result := C_GVVersionSt +  ' - © Gilles Vasseur';
end;

function TGVTransition.GetOpacity(Up: Boolean): Integer;
// *** calcul de l'opacité en fonction du pourcentage effectué ***
begin
  Result := ifthen(Up, fStep * C_MaxOpacity div fLastLoop, C_MaxOpacity -
    fStep * C_MaxOpacity div fLastLoop);
end;

function TGVTransition.NumberOfStrips(ASize: Integer): Integer;
// *** calcul automatique du nombre de bandes ***
begin
  if ASize < C_MinComputeSize then
    Result := 1
  else
  if ASize < (C_MaxComputeSize div 5) then
    Result := 5
  else
  if ASize < (C_MaxComputeSize div 2) then
    Result := 10
  else
  if ASize < C_MaxComputeSize then
    Result := ASize div C_DefaultStripWidth
  else
    Result := 30;
end;

procedure TGVTransition.Start;
// *** déclenchement de la transition ***
begin
  case fState of
    trsRunning: exit;
    trsSuspended: State := trsRestarting;
  else
    fStep := 0;
    InitTransition;
    State := trsStarting;
  end;
  State := trsRunning;
end;

procedure TGVTransition.Stop;
// *** arrêt d'une transition ***
begin
  if State = trsRunning then
  begin
    State := trsStopped;
    fStep := fLastLoop;
    State := trsWaiting;
  end;
end;

procedure TGVTransition.Reset;
// *** remise à zéro de l'affichage ***
begin
  DoReset;
  Repaint;
end;

procedure TGVTransition.Suspend;
// *** suspension de la transition ***
begin
  if fState = trsRunning then
    State := trsSuspended;
end;

procedure TGVTransition.Resume;
// reprise d'une animation ***
begin
  if fState = trsSuspended then
    Start;
end;

function TGVTransition.IsRunning: Boolean;
// *** transition en cours ? ***
begin
  Result := (fState = trsRunning);
end;

procedure TGVTransition.DoNegateMask;
// *** masque inversé ***
begin
  fLBGRAMask.Negative;
end;

initialization
  {$I transitions.lrs}
end.
